'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var i18next = _interopDefault(require('i18next'));
var BabelTypes = require('@babel/types');
var fs = _interopDefault(require('fs'));
var path = _interopDefault(require('path'));

const COMMENT_HINT_PREFIX = 'i18next-extract-';
const COMMENT_HINTS_KEYWORDS = {
    DISABLE: {
        LINE: COMMENT_HINT_PREFIX + 'disable-line',
        NEXT_LINE: COMMENT_HINT_PREFIX + 'disable-next-line',
        SECTION_START: COMMENT_HINT_PREFIX + 'disable',
        SECTION_STOP: COMMENT_HINT_PREFIX + 'enable',
    },
    NAMESPACE: {
        LINE: COMMENT_HINT_PREFIX + 'mark-ns-line',
        NEXT_LINE: COMMENT_HINT_PREFIX + 'mark-ns-next-line',
        SECTION_START: COMMENT_HINT_PREFIX + 'mark-ns-start',
        SECTION_STOP: COMMENT_HINT_PREFIX + 'mark-ns-stop',
    },
    CONTEXT: {
        LINE: COMMENT_HINT_PREFIX + 'mark-context-line',
        NEXT_LINE: COMMENT_HINT_PREFIX + 'mark-context-next-line',
        SECTION_START: COMMENT_HINT_PREFIX + 'mark-context-start',
        SECTION_STOP: COMMENT_HINT_PREFIX + 'mark-context-stop',
    },
    PLURAL: {
        LINE: COMMENT_HINT_PREFIX + 'mark-plural-line',
        NEXT_LINE: COMMENT_HINT_PREFIX + 'mark-plural-next-line',
        SECTION_START: COMMENT_HINT_PREFIX + 'mark-plural-start',
        SECTION_STOP: COMMENT_HINT_PREFIX + 'mark-plural-stop',
    },
};
/**
 * Given a Babel BaseComment, extract base comment hints.
 * @param baseComment babel comment
 * @yields Comment hint without line interval information.
 */
function* extractCommentHintFromBaseComment(baseComment) {
    for (const line of baseComment.value.split(/\r?\n/)) {
        const trimmedValue = line.trim();
        const keyword = trimmedValue.split(/\s+/)[0];
        const value = trimmedValue.split(/\s+(.+)/)[1] || '';
        for (let [commentHintType, commentHintKeywords] of Object.entries(COMMENT_HINTS_KEYWORDS)) {
            for (let [commentHintScope, commentHintKeyword] of Object.entries(commentHintKeywords)) {
                if (keyword === commentHintKeyword) {
                    yield {
                        type: commentHintType,
                        scope: commentHintScope,
                        value,
                        baseComment: baseComment,
                    };
                }
            }
        }
    }
}
/**
 * Given an array of comment hints, compute their intervals.
 * @param commentHints comment hints without line intervals information.
 * @returns Comment hints with line interval information.
 */
function computeCommentHintsIntervals(commentHints) {
    const result = Array();
    for (const commentHint of commentHints) {
        if (commentHint.scope === 'LINE') {
            result.push({
                startLine: commentHint.baseComment.loc.start.line,
                stopLine: commentHint.baseComment.loc.start.line,
                ...commentHint,
            });
        }
        if (commentHint.scope === 'NEXT_LINE') {
            result.push({
                startLine: commentHint.baseComment.loc.end.line + 1,
                stopLine: commentHint.baseComment.loc.end.line + 1,
                ...commentHint,
            });
        }
        if (commentHint.scope === 'SECTION_START') {
            result.push({
                startLine: commentHint.baseComment.loc.start.line,
                stopLine: Infinity,
                ...commentHint,
            });
        }
        if (commentHint.scope === 'SECTION_STOP') {
            for (const res of result) {
                if (res.type === commentHint.type &&
                    res.scope === 'SECTION_START' &&
                    res.stopLine === Infinity) {
                    res.stopLine = commentHint.baseComment.loc.start.line;
                }
            }
        }
    }
    return result;
}
/**
 * Given Babel comments, extract the comment hints.
 * @param baseComments Babel comments (ordered by line)
 */
function parseCommentHints(baseComments) {
    const baseCommentHints = Array();
    for (const baseComment of baseComments) {
        baseCommentHints.push(...extractCommentHintFromBaseComment(baseComment));
    }
    return computeCommentHintsIntervals(baseCommentHints);
}
/**
 * Find comment hint of a given type that applies to a Babel node path.
 * @param path babel node path
 * @param commentHintType Type of comment hint to look for.
 * @param commentHints All the comment hints, as returned by parseCommentHints function.
 */
function getCommentHintForPath(path, commentHintType, commentHints) {
    if (!path.node.loc)
        return null;
    const nodeLine = path.node.loc.start.line;
    for (const commentHint of commentHints) {
        if (commentHint.type === commentHintType &&
            commentHint.startLine <= nodeLine &&
            nodeLine <= commentHint.stopLine) {
            return commentHint;
        }
    }
    return null;
}

/**
 * Error thrown in case extraction of a node failed.
 */
class ExtractionError extends Error {
}
/**
 * Given a value, if the value is an array, return the first
 * item of the array. Otherwise, return the value.
 *
 * This is mainly useful to parse namespaces which can be strings
 * as well as array of strings.
 */
function getFirstOrNull(val) {
    if (Array.isArray(val))
        val = val[0];
    return val === undefined ? null : val;
}
/**
 * Given comment hints and a path, infer every I18NextOption we can from the comment hints.
 * @param path path on which the comment hints should apply
 * @param commentHints parsed comment hints
 * @returns every parsed option that could be infered.
 */
function parseI18NextOptionsFromCommentHints(path, commentHints) {
    const nsCommentHint = getCommentHintForPath(path, 'NAMESPACE', commentHints);
    const contextCommentHint = getCommentHintForPath(path, 'CONTEXT', commentHints);
    const pluralCommentHint = getCommentHintForPath(path, 'PLURAL', commentHints);
    const res = {};
    if (nsCommentHint !== null) {
        res.ns = nsCommentHint.value;
    }
    if (contextCommentHint !== null) {
        if (['', 'enable'].includes(contextCommentHint.value)) {
            res.contexts = true;
        }
        else if (contextCommentHint.value === 'disable') {
            res.contexts = false;
        }
        else {
            try {
                let val = JSON.parse(contextCommentHint.value);
                if (Array.isArray(val))
                    res.contexts = val;
                else
                    res.contexts = [contextCommentHint.value];
            }
            catch (err) {
                res.contexts = [contextCommentHint.value];
            }
        }
    }
    if (pluralCommentHint !== null) {
        if (pluralCommentHint.value === 'disable') {
            res.hasCount = false;
        }
        else {
            res.hasCount = true;
        }
    }
    return res;
}
/**
 * Improved version of BabelCore `referencesImport` function that also tries to detect wildcard
 * imports.
 */
function referencesImport(nodePath, moduleSource, importName) {
    if (nodePath.referencesImport(moduleSource, importName))
        return true;
    if (nodePath.isMemberExpression() || nodePath.isJSXMemberExpression()) {
        const obj = nodePath.get('object');
        const prop = nodePath.get('property');
        if (Array.isArray(prop) ||
            (!prop.isIdentifier() && !prop.isJSXIdentifier()))
            return false;
        return (obj.referencesImport(moduleSource, '*') && prop.node.name === importName);
    }
    return false;
}
/**
 * Evaluates a node path if it can be evaluated with confidence.
 *
 * @param path: node path to evaluate
 * @returns null if the node path couldn't be evaluated
 */
function evaluateIfConfident(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
path) {
    if (!path || !path.node) {
        return null;
    }
    const evaluation = path.evaluate();
    if (evaluation.confident) {
        return evaluation.value;
    }
    return null;
}
/**
 * Generator that iterates on all keys in an object expression.
 * @param path the node path of the object expression
 * @param key the key to find in the object expression.
 * @yields [evaluated key, node path of the object expression property]
 */
function* iterateObjectExpression(path) {
    const properties = path.get('properties');
    for (const prop of properties) {
        const keyPath = prop.get('key');
        if (Array.isArray(keyPath))
            continue;
        let keyEvaluation = null;
        if (keyPath.isLiteral()) {
            keyEvaluation = evaluateIfConfident(keyPath);
        }
        else if (keyPath.isIdentifier()) {
            keyEvaluation = keyPath.node.name;
        }
        else {
            continue;
        }
        yield [keyEvaluation, prop];
    }
}
/**
 * Try to find a key in an object expression.
 * @param path the node path of the object expression
 * @param key the key to find in the object expression.
 * @returns the corresponding node or null if it wasn't found
 */
function findKeyInObjectExpression(path, key) {
    for (const [keyEvaluation, prop] of iterateObjectExpression(path)) {
        if (keyEvaluation === key)
            return prop;
    }
    return null;
}
/**
 * Find a JSX attribute given its name.
 * @param path path of the jsx attribute
 * @param name name of the attribute to look for
 * @return The JSX attribute corresponding to the given name, or null if no
 *   attribute with this name could be found.
 */
function findJSXAttributeByName(path, name) {
    const openingElement = path.get('openingElement');
    const attributes = openingElement.get('attributes');
    for (const attribute of attributes) {
        if (!attribute.isJSXAttribute())
            continue;
        const attributeName = attribute.get('name');
        if (!attributeName.isJSXIdentifier())
            continue;
        if (name === attributeName.node.name)
            return attribute;
    }
    return null;
}

/**
 * Check whether a given CallExpression path is a global call to `i18next.t`
 * function.
 * @param path: node path to check
 * @param config: plugin configuration
 * @returns true if the given call expression is indeed a call to i18next.t.
 */
function isSimpleTCall(path, config) {
    const callee = path.get('callee');
    if (!callee.isMemberExpression())
        return false;
    const obj = callee.get('object');
    if (!obj.isIdentifier())
        return false;
    const prop = callee.get('property');
    if (Array.isArray(prop) || !prop.isIdentifier())
        return false;
    return (config.i18nextInstanceNames.includes(obj.node.name) &&
        prop.node.name === 't');
}
/**
 * Parse options of a `i18next.t(…)` call.
 * @param path: NodePath representing the second argument of the `t()` call
 *   (i.e. the i18next options)
 * @returns an object indicating whether the parsed options have context
 *   and/or count.
 */
function parseTCallOptions(path) {
    const res = {
        contexts: false,
        hasCount: false,
        ns: null,
    };
    if (!path)
        return res;
    // Try brutal evaluation first.
    const optsEvaluation = evaluateIfConfident(path);
    if (optsEvaluation !== null) {
        res.contexts = 'context' in optsEvaluation;
        res.hasCount = 'count' in optsEvaluation;
        const evaluatedNamespace = optsEvaluation['ns'];
        res.ns = getFirstOrNull(evaluatedNamespace);
    }
    else if (path.isObjectExpression()) {
        // It didn't work. Let's try to parse object expression keys.
        res.contexts = findKeyInObjectExpression(path, 'context') !== null;
        res.hasCount = findKeyInObjectExpression(path, 'count') !== null;
        const nsNode = findKeyInObjectExpression(path, 'ns');
        const nsNodeEvaluation = evaluateIfConfident(nsNode);
        res.ns = getFirstOrNull(nsNodeEvaluation);
    }
    return res;
}
/**
 * Given a call to `i18next.t`, find the key and the options.
 *
 * @param path NodePath of the `i18next.t` call.
 * @param commentHints parsed comment hints
 * @throws ExtractionError when the extraction failed for the `t` call.
 */
function extractTCall(path, commentHints) {
    const args = path.get('arguments');
    const keyEvaluation = evaluateIfConfident(args[0]);
    if (typeof keyEvaluation !== 'string') {
        throw new ExtractionError(`Couldn't evaluate i18next key. You should either make the key ` +
            `evaluable or skip the line using a skip comment (/* ` +
            `${COMMENT_HINTS_KEYWORDS.DISABLE.LINE} */ or /* ` +
            `${COMMENT_HINTS_KEYWORDS.DISABLE.NEXT_LINE} */).`);
    }
    return {
        key: keyEvaluation,
        parsedOptions: {
            ...parseTCallOptions(args[1]),
            ...parseI18NextOptionsFromCommentHints(path, commentHints),
        },
        nodePath: path,
    };
}
/**
 * Parse a call expression (likely `i18next.t`) to find its translation keys
 * and i18next options.
 *
 * @param path: node path of the t function call.
 * @param config: plugin configuration
 * @param commentHints: parsed comment hints
 * @param skipCheck: set to true if you know that the call expression arguments
 *   already is a `t` function.
 */
function extractTFunction(path, config, commentHints = [], skipCheck = false) {
    if (getCommentHintForPath(path, 'DISABLE', commentHints))
        return [];
    if (!skipCheck && !isSimpleTCall(path, config))
        return [];
    return [extractTCall(path, commentHints)];
}

/**
 * Check whether a given CallExpression path is a call to `useTranslation` hook.
 * @param path: node path to check
 * @returns true if the given call expression is indeed a call to
 *   `useTranslation`
 */
function isUseTranslationHook(path) {
    const callee = path.get('callee');
    return referencesImport(callee, 'react-i18next', 'useTranslation');
}
/**
 * Parse `useTranslation()` hook to extract all its translation keys and
 * options.
 * @param path: useTranslation call node path.
 * @param config: plugin configuration
 * @param commentHints: parsed comment hints
 */
function extractUseTranslationHook(path, config, commentHints = []) {
    if (!isUseTranslationHook(path))
        return [];
    let ns;
    const nsCommentHint = getCommentHintForPath(path, 'NAMESPACE', commentHints);
    if (nsCommentHint) {
        // We got a comment hint, take its value as namespace.
        ns = nsCommentHint.value;
    }
    else {
        // Otherwise, try to get namespace from arguments.
        const namespaceArgument = path.get('arguments')[0];
        ns = getFirstOrNull(evaluateIfConfident(namespaceArgument));
    }
    const parentPath = path.parentPath;
    if (!parentPath.isVariableDeclarator())
        return [];
    const id = parentPath.get('id');
    const tBinding = id.scope.bindings['t'];
    if (!tBinding)
        return [];
    let keys = Array();
    for (const reference of tBinding.referencePaths) {
        if (reference.parentPath.isCallExpression()) {
            keys = [
                ...keys,
                ...extractTFunction(reference.parentPath, config, commentHints, true).map(k => ({
                    // Add namespace if it was not explicitely set in t() call.
                    ...k,
                    parsedOptions: {
                        ...k.parsedOptions,
                        ns: k.parsedOptions.ns || ns,
                    },
                })),
            ];
        }
    }
    return keys;
}

/**
 * Check whether a given JSXElement is a Translation render prop.
 * @param path: node path to check
 * @returns true if the given element is indeed a `Translation` render prop.
 */
function isTranslationRenderProp(path) {
    const openingElement = path.get('openingElement');
    return referencesImport(openingElement.get('name'), 'react-i18next', 'Translation');
}
/**
 * Parse `Translation` render prop to extract all its translation keys and
 * options.
 *
 * @param path: node path of Translation JSX element.
 * @param config: plugin configuration
 * @param commentHints: parsed comment hints
 */
function extractTranslationRenderProp(path, config, commentHints = []) {
    if (!isTranslationRenderProp(path))
        return [];
    let ns;
    const nsCommentHint = getCommentHintForPath(path, 'NAMESPACE', commentHints);
    if (nsCommentHint) {
        // We got a comment hint, take its value as namespace.
        ns = nsCommentHint.value;
    }
    else {
        // Try to parse ns property
        const nsAttr = findJSXAttributeByName(path, 'ns');
        if (nsAttr) {
            let value = nsAttr.get('value');
            if (value.isJSXExpressionContainer())
                value = value.get('expression');
            ns = getFirstOrNull(evaluateIfConfident(value));
        }
    }
    // We expect at least "<Translation>{(t) => …}</Translation>
    const expressionContainer = path
        .get('children')
        .filter(p => p.isJSXExpressionContainer())[0];
    if (!expressionContainer || !expressionContainer.isJSXExpressionContainer())
        return [];
    const expression = expressionContainer.get('expression');
    if (!expression.isArrowFunctionExpression())
        return [];
    const tParam = expression.get('params')[0];
    if (!tParam)
        return [];
    const tBinding = tParam.scope.bindings['t'];
    if (!tBinding)
        return [];
    let keys = Array();
    for (const reference of tBinding.referencePaths) {
        if (reference.parentPath.isCallExpression()) {
            keys = [
                ...keys,
                ...extractTFunction(reference.parentPath, config, commentHints, true).map(k => ({
                    // Add namespace if it was not explicitely set in t() call.
                    ...k,
                    parsedOptions: {
                        ...k.parsedOptions,
                        ns: k.parsedOptions.ns || ns,
                    },
                })),
            ];
        }
    }
    return keys;
}

/**
 * Check whether a given JSXElement is a Trans component.
 * @param path: node path to check
 * @returns true if the given element is indeed a `Trans` component.
 */
function isTransComponent(path) {
    const openingElement = path.get('openingElement');
    return referencesImport(openingElement.get('name'), 'react-i18next', 'Trans');
}
/**
 * Given a Trans component, extract its options.
 * @param path The node path of the JSX Element of the trans component
 * @param commentHints Parsed comment hints.
 * @returns The parsed i18next options
 */
function parseTransComponentOptions(path, commentHints) {
    const res = {
        contexts: false,
        hasCount: false,
        ns: null,
    };
    const countAttr = findJSXAttributeByName(path, 'count');
    res.hasCount = countAttr !== null;
    const tOptionsAttr = findJSXAttributeByName(path, 'tOptions');
    if (tOptionsAttr) {
        const value = tOptionsAttr.get('value');
        if (value.isJSXExpressionContainer()) {
            const expression = value.get('expression');
            if (expression.isObjectExpression()) {
                res.contexts =
                    findKeyInObjectExpression(expression, 'context') !== null;
            }
        }
    }
    const nsAttr = findJSXAttributeByName(path, 'ns');
    if (nsAttr) {
        let value = nsAttr.get('value');
        if (value.isJSXExpressionContainer())
            value = value.get('expression');
        res.ns = getFirstOrNull(evaluateIfConfident(value));
    }
    return {
        ...res,
        ...parseI18NextOptionsFromCommentHints(path, commentHints),
    };
}
/**
 * Given the node path of a Trans component, try to extract its key from its
 *   attributes.
 * @param path node path of the Trans component.
 * @returns the component key if it was found.
 * @throws ExtractionError if the i18nKey attribute was present but not
 *   evaluable.
 */
function parseTransComponentKeyFromAttributes(path) {
    const error = new ExtractionError(`Couldn't evaluate i18next key in Trans component. You should either ` +
        `make the i18nKey attribute evaluable or skip the line using a skip ` +
        `comment (/* ${COMMENT_HINTS_KEYWORDS.DISABLE.LINE} */ or /* ` +
        `${COMMENT_HINTS_KEYWORDS.DISABLE.NEXT_LINE} */).`);
    const keyAttribute = findJSXAttributeByName(path, 'i18nKey');
    if (!keyAttribute)
        return null;
    const keyAttributeValue = keyAttribute.get('value');
    const keyEvaluation = evaluateIfConfident(keyAttributeValue.isJSXExpressionContainer()
        ? keyAttributeValue.get('expression')
        : keyAttributeValue);
    if (typeof keyEvaluation !== 'string') {
        throw error;
    }
    return keyEvaluation;
}
/**
 * Given the node path of a Trans component, try to extract its key from its
 *   children.
 * @param path node path of the Trans component.
 * @returns the component key if it was found.
 * @throws ExtractionError if the extraction did not succeed.
 */
function parseTransComponentKeyFromChildren(path) {
    const error = new ExtractionError(`Couldn't evaluate i18next key in Trans component. You should either ` +
        `set the i18nKey attribute to an evaluable value, or make the Trans ` +
        `component content evaluable or skip the line using a skip comment ` +
        `(/* ${COMMENT_HINTS_KEYWORDS.DISABLE.LINE} */ or /* ` +
        `${COMMENT_HINTS_KEYWORDS.DISABLE.NEXT_LINE} */).`);
    let children = path.get('children');
    let result = '';
    // Remove trailing empty JSXText nodes. They should not be taken into
    // account for the indices.
    const firstChild = children[0];
    if (firstChild &&
        firstChild.isJSXText() &&
        firstChild.node.value.trim() === '') {
        children = children.slice(1);
    }
    const lastChild = children[children.length - 1];
    if (lastChild &&
        lastChild.isJSXText() &&
        lastChild.node.value.trim() === '') {
        children = children.slice(0, children.length - 1);
    }
    // Filter out empty containers. They do not affect indices.
    children = children.filter(p => {
        if (!p.isJSXExpressionContainer())
            return true;
        const expr = p.get('expression');
        return !expr.isJSXEmptyExpression();
    });
    // We can then iterate on the children.
    for (let [i, child] of children.entries()) {
        if (child.isJSXExpressionContainer()) {
            // We have an expression container: {…}
            const expression = child.get('expression');
            const evaluation = evaluateIfConfident(expression);
            if (evaluation !== null && typeof evaluation === 'string') {
                // We have an evaluable JSX expression like {'hello'}
                result += evaluation.toString();
                continue;
            }
            if (expression.isObjectExpression()) {
                // We have an expression like {{name}} or {{name: userName}}
                const it = iterateObjectExpression(expression);
                const key0 = it.next().value;
                if (!key0 || !it.next().done) {
                    // Probably got empty object expression like {{}}
                    // or {{foo,bar}}
                    throw error;
                }
                result += `{{${key0[0]}}}`;
                continue;
            }
            if (expression.isIdentifier()) {
                // We have an identifier like {myPartialComponent}
                // We try to find the latest declaration and substitute the identifier.
                const bindings = expression.scope.bindings[expression.node.name];
                if (!bindings)
                    throw error;
                const declarationExpressions = [
                    ...(bindings.path.isVariableDeclarator()
                        ? [bindings.path.get('init')]
                        : []),
                    ...bindings.constantViolations
                        .filter(p => p.isAssignmentExpression())
                        .map(p => p.get('right')),
                ];
                if (declarationExpressions.length === 0)
                    throw error;
                const latestDeclarator = declarationExpressions[declarationExpressions.length - 1];
                if (Array.isArray(latestDeclarator))
                    throw error;
                const evaluation = evaluateIfConfident(latestDeclarator);
                if (evaluation !== null) {
                    // It could be evaluated, it's probably something like 'hello'
                    result += evaluation;
                    continue;
                }
                else if (latestDeclarator.isJSXElement()) {
                    // It's a JSX element. Let's act as if it was inline and move along.
                    child = latestDeclarator;
                }
                else {
                    throw error;
                }
            }
        }
        if (child.isJSXText()) {
            // Simple JSX text.
            result +=
                // Let's sanitize the value a bit.
                child.node.value
                    // Strip line returns at start
                    .replace(/^\s*(\r?\n)+\s*/gm, '')
                    // Strip line returns at end
                    .replace(/\s*(\r?\n)+\s*$/gm, '')
                    // Replace other line returns with one space
                    .replace(/\s*(\r?\n)+\s*/gm, ' ');
            continue;
        }
        if (child.isJSXElement()) {
            // got a JSX element.
            if (child.get('closingElement').node === null) {
                // opening tag without closing tag (e.g. <br />)
                result += `<${i}>`;
                continue;
            }
            // it's nested. let's recurse.
            result += `<${i}>${parseTransComponentKeyFromChildren(child)}</${i}>`;
            continue;
        }
    }
    return result;
}
/**
 * Parse `Trans` component to extract all its translation keys and i18next
 * options.
 *
 * @param path: node path of Trans JSX element.
 * @param config: plugin configuration
 * @param commentHints: parsed comment hints
 */
function extractTransComponent(path, config, commentHints = []) {
    if (getCommentHintForPath(path, 'DISABLE', commentHints))
        return [];
    if (!isTransComponent(path))
        return [];
    const keyEvaluation = parseTransComponentKeyFromAttributes(path) ||
        parseTransComponentKeyFromChildren(path);
    const parsedOptions = parseTransComponentOptions(path, commentHints);
    return [
        {
            key: keyEvaluation,
            parsedOptions,
            nodePath: path,
        },
    ];
}

/**
 * Parse namespace and key path from an extracted key.
 * @param key: key to parse
 * @param config: plugin configuration
 */
function parseExtractedKey(key, config) {
    let cleanKey = key.key;
    let ns = key.parsedOptions.ns || config.defaultNS;
    if (config.nsSeparator) {
        const nsSeparatorPos = cleanKey.indexOf(config.nsSeparator);
        if (nsSeparatorPos !== -1) {
            ns = cleanKey.slice(0, nsSeparatorPos);
            cleanKey = cleanKey.slice(nsSeparatorPos + 1);
        }
    }
    let keyPath = Array();
    if (config.keySeparator) {
        const fullPath = cleanKey.split(config.keySeparator);
        keyPath = fullPath.slice(0, fullPath.length - 1);
        cleanKey = fullPath[fullPath.length - 1];
    }
    return {
        ...key,
        cleanKey,
        keyPath,
        ns,
        isDerivedKey: false,
    };
}
/**
 * Compute all derived keys for a local from a key and parsed i18next options.
 *
 * e.g.
 *   ({'foo', {contexts: false, hasCount: true}}, 'en')
 *     => ['foo', 'foo_plural']
 *   ({'bar', {contexts: ['male', 'female'], hasCount: true}}, 'en')
 *     => ['foo_male', 'foo_male_plural', 'foo_female', 'foo_female_plural']
 *
 * @param extractedKey key that was extracted with an extractor.
 * @param locale locale code
 * @returns All derived keys that could be found from TranslationKey for
 *   locale.
 */
function computeDerivedKeys(extractedKey, locale, config) {
    const translationKey = parseExtractedKey(extractedKey, config);
    const { parsedOptions, cleanKey: key } = translationKey;
    let keys = [translationKey];
    if (parsedOptions.contexts !== false) {
        // Add all context suffixes
        // For instance, if key is "foo", may want
        // ["foo", "foo_male", "foo_female"] depending on defaultContexts value.
        const contexts = Array.isArray(parsedOptions.contexts)
            ? parsedOptions.contexts
            : config.defaultContexts;
        keys = contexts.map(v => {
            if (v === '')
                return translationKey;
            return {
                ...translationKey,
                cleanKey: key + config.contextSeparator + v,
                isDerivedKey: true,
            };
        });
    }
    if (parsedOptions.hasCount) {
        // See https://www.i18next.com/translation-function/plurals#how-to-find-the-correct-plural-suffix
        const pluralRule = i18next.services.pluralResolver.getRule(locale);
        const numberOfPlurals = pluralRule.numbers.length;
        if (numberOfPlurals === 1) {
            keys = keys.map(k => ({
                ...k,
                cleanKey: k.cleanKey + config.pluralSeparator + '0',
                isDerivedKey: true,
            }));
        }
        else if (numberOfPlurals === 2) {
            keys = keys.reduce((accumulator, k) => [
                ...accumulator,
                k,
                {
                    ...k,
                    cleanKey: k.cleanKey + config.pluralSeparator + 'plural',
                    isDerivedKey: true,
                },
            ], Array());
        }
        else {
            keys = keys.reduce((accumulator, k) => [
                ...accumulator,
                ...Array(numberOfPlurals)
                    .fill(null)
                    .map((_, idx) => ({
                    ...k,
                    cleanKey: k.cleanKey + config.pluralSeparator + idx,
                    isDerivedKey: true,
                })),
            ], Array());
        }
    }
    return keys;
}

function coalesce(v, defaultVal) {
    return v === undefined ? defaultVal : v;
}
/**
 * Given Babel options, return an initialized Config object.
 *
 * @param opts plugin options given by Babel
 */
function parseConfig(opts) {
    return {
        locales: coalesce(opts.locales, ['en']),
        defaultNS: coalesce(opts.defaultNS, 'translation'),
        pluralSeparator: coalesce(opts.pluralSeparator, '_'),
        contextSeparator: coalesce(opts.contextSeparator, '_'),
        keySeparator: coalesce(opts.keySeparator, '.'),
        nsSeparator: coalesce(opts.nsSeparator, ':'),
        i18nextInstanceNames: coalesce(opts.i18nextInstanceNames, [
            'i18next',
            'i18n',
        ]),
        defaultContexts: coalesce(opts.defaultContexts, ['', 'male', 'female']),
        outputPath: coalesce(opts.outputPath, './extractedTranslations/{{locale}}/{{ns}}.json'),
        defaultValue: coalesce(opts.defaultValue, ''),
        keyAsDefaultValue: coalesce(opts.keyAsDefaultValue, false),
        keyAsDefaultValueForDerivedKeys: coalesce(opts.keyAsDefaultValueForDerivedKeys, true),
        discardOldKeys: coalesce(opts.discardOldKeys, false),
        jsonSpace: coalesce(opts.jsonSpace, 2),
    };
}

const PLUGIN_NAME = 'babel-plugin-i18next-extract';

/**
 * Thrown when an error happens during the translations export.
 */
class ExportError extends Error {
}
/**
 * This creates a new empty cache for the exporter.
 *
 * The cache is required by the exporter and is used to merge the translations
 * from the original translation file. It will be  mutated by the exporter
 * and the same instance must be given untouched across export calls.
 */
function createExporterCache() {
    return {
        originalTranslationFiles: {},
        currentTranslationFiles: {},
    };
}
/**
 * Take a deep translation file and flatten it.
 * This is to simplify merging keys later on.
 *
 * @param deep Deep translation file.
 */
function flattenTranslationFile(deep) {
    const result = {};
    for (const [k, v] of Object.entries(deep)) {
        if (typeof v === 'object' && v !== null) {
            // Nested case, we must recurse
            const flat = flattenTranslationFile(v);
            for (const [flatK, flatV] of Object.entries(flat)) {
                result[JSON.stringify([k, ...JSON.parse(flatK)])] = flatV;
            }
        }
        else {
            // Leaf, just set the value.
            result[JSON.stringify([k])] = v;
        }
    }
    return result;
}
/**
 * Take a flat translation file and make it deep.
 *
 * This is to make a flat translation file ready to be exported to a file.
 *
 * @param flat Flat translation file as returned by flattenTranslationFile
 */
function unflattenTranslationFile(flat) {
    const result = {};
    for (const [k, v] of Object.entries(flat)) {
        const keyPath = JSON.parse(k);
        const cleanKey = keyPath.pop();
        const error = new ExportError(`${PLUGIN_NAME}: Couldn't export translations. Key "${keyPath}" ` +
            `has a conflict.`);
        let obj = result;
        for (const p of keyPath) {
            const currentValue = obj[p];
            if (typeof currentValue === 'string' || currentValue === null) {
                throw error;
            }
            obj = obj[p] = currentValue || {};
        }
        const currentValue = obj[cleanKey];
        if (typeof currentValue === 'object' && currentValue !== null) {
            throw error;
        }
        obj[cleanKey] = v;
    }
    return result;
}
/**
 * Load translation given a file. If the file is not found, default to empty
 * object.
 * @param filePath Path of the JSON translation file to load.
 */
function loadTranslationFile(filePath) {
    let content;
    try {
        content = fs.readFileSync(filePath, { encoding: 'utf8' });
    }
    catch (err) {
        if (err.code === 'ENOENT')
            return {};
        throw err;
    }
    return flattenTranslationFile(JSON.parse(content));
}
/**
 * Create a new translationFile with a key added to it.
 *
 * @param translationFile The translation file to add the key to.
 * @param key The translation key to add.
 * @param locale Current locale being exported
 * @param config Configuration (that will help setting the proper default
 *   value)
 */
function addKeyToTranslationFile(translationFile, key, locale, config) {
    // compute the default value
    let defaultValue = config.defaultValue;
    const keyAsDefaultValueEnabled = config.keyAsDefaultValue === true ||
        (Array.isArray(config.keyAsDefaultValue) &&
            config.keyAsDefaultValue.includes(locale));
    const keyAsDefaultValueForDerivedKeys = config.keyAsDefaultValueForDerivedKeys;
    if (keyAsDefaultValueEnabled &&
        (keyAsDefaultValueForDerivedKeys || !key.isDerivedKey)) {
        defaultValue = key.cleanKey;
    }
    return {
        [JSON.stringify([...key.keyPath, key.cleanKey])]: defaultValue,
        ...translationFile,
    };
}
/**
 * Exports all given translation keys as JSON.
 *
 * @param keys: translation keys to export
 * @param locale: the locale to export
 * @param config: plugin configuration
 * @param cache: cache instance to use (see createExporterCache)
 */
function exportTranslationKeys(keys, locale, config, cache) {
    const keysPerFilepath = {};
    for (const key of keys) {
        // Figure out in which path each key should go.
        const filePath = config.outputPath
            .replace('{{locale}}', locale)
            .replace('{{ns}}', key.ns);
        keysPerFilepath[filePath] = [...(keysPerFilepath[filePath] || []), key];
    }
    for (const [filePath, keysForFilepath] of Object.entries(keysPerFilepath)) {
        if (!(filePath in cache.originalTranslationFiles)) {
            // Cache original translation file so that we don't loose it across babel
            // passes.
            cache.originalTranslationFiles[filePath] = loadTranslationFile(filePath);
        }
        let translationFile = cache.currentTranslationFiles[filePath] || {};
        for (const k of keysForFilepath) {
            translationFile = addKeyToTranslationFile(translationFile, k, locale, config);
        }
        cache.currentTranslationFiles[filePath] = translationFile;
        let deepTranslationFile;
        if (config.discardOldKeys) {
            const originalTranslationFile = cache.originalTranslationFiles[filePath];
            const alreadyTranslated = Object.keys(originalTranslationFile)
                .filter(k => k in translationFile)
                .reduce((accumulator, k) => ({
                ...accumulator,
                [k]: originalTranslationFile[k],
            }), {});
            deepTranslationFile = unflattenTranslationFile({
                ...translationFile,
                ...alreadyTranslated,
            });
        }
        else {
            deepTranslationFile = unflattenTranslationFile({
                ...translationFile,
                ...cache.originalTranslationFiles[filePath],
            });
        }
        // Finally do the export
        const directoryPath = path.dirname(filePath);
        fs.mkdirSync(directoryPath, { recursive: true });
        fs.writeFileSync(filePath, JSON.stringify(deepTranslationFile, null, config.jsonSpace), {
            encoding: 'utf8',
        });
    }
}

/**
 * Find whether a given function or class is wrapped with "withTranslation" HOC
 * somewhere.
 * @param path Function or class declaration node path.
 * @returns "withTranslation()()" call expression if found. Else null.
 */
function findWithTranslationHOCCallExpression(path) {
    let functionIdentifier = path.get('id');
    if (!Array.isArray(functionIdentifier) &&
        !functionIdentifier.isIdentifier() &&
        path.parentPath.isVariableDeclarator()) {
        // It doesn't look like "function MyComponent(…)"
        // but like be "const MyComponent = (…) => …" or "const MyComponent = function(…) { … }"
        functionIdentifier = path.parentPath.get('id');
    }
    if (Array.isArray(functionIdentifier) || !functionIdentifier.isIdentifier())
        return null;
    // Try to find a withTranslation() call in parent scope
    for (const refPath of path.parentPath.scope.bindings[functionIdentifier.node.name].referencePaths) {
        const callExpr = refPath.findParent(parentPath => {
            if (!parentPath.isCallExpression())
                return false;
            const callee = parentPath.get('callee');
            return (callee.isCallExpression() &&
                referencesImport(callee.get('callee'), 'react-i18next', 'withTranslation'));
        });
        if (callExpr !== null) {
            return callExpr.get('callee');
        }
    }
    return null;
}
/**
 * Try to find "t" in an object spread. Useful when looking for the "t" key
 * in a spread object. e.g. const {t} = props;
 *
 * @param path object pattern
 * @returns t identifier or null of it was not found in the object pattern.
 */
function findTFunctionIdentifierInObjectPattern(path) {
    const props = path.get('properties');
    for (const prop of props) {
        if (prop.isObjectProperty()) {
            const key = prop.get('key');
            if (!Array.isArray(key) && key.isIdentifier() && key.node.name === 't') {
                return key;
            }
        }
    }
    return null;
}
/**
 * Find T function calls from a props assignment. Prop assignment can occur
 * in function parameters (i.e. "function Component(props)" or
 * "function Component({t})") or in a variable declarator (i.e.
 * "const props = …" or "const {t} = props").
 *
 * @param propsId identifier for the prop assignment. e.g. "props" or "{t}"
 * @returns Call expressions to t function.
 */
function findTFunctionCallsFromPropsAssignment(propsId) {
    const tReferences = Array();
    if (propsId.isObjectPattern()) {
        // got "function MyComponent({t, other, props})"
        // or "const {t, other, props} = this.props"
        // we want to find references to "t"
        const tFunctionIdentifier = findTFunctionIdentifierInObjectPattern(propsId);
        if (tFunctionIdentifier === null)
            return [];
        const tBinding = tFunctionIdentifier.scope.bindings[tFunctionIdentifier.node.name];
        if (!tBinding)
            return [];
        tReferences.push(...tBinding.referencePaths);
    }
    else if (propsId.isIdentifier()) {
        // got "function MyComponent(props)"
        // or "const props = this.props"
        // we want to find references to props.t
        const references = propsId.scope.bindings[propsId.node.name].referencePaths;
        for (const reference of references) {
            if (reference.parentPath.isMemberExpression()) {
                const prop = reference.parentPath.get('property');
                if (!Array.isArray(prop) && prop.node.name === 't') {
                    tReferences.push(reference.parentPath);
                }
            }
        }
    }
    // We have candidates. Let's see if t references are actual calls to the t
    // function
    const tCalls = Array();
    for (const tCall of tReferences) {
        if (tCall.parentPath.isCallExpression()) {
            tCalls.push(tCall.parentPath);
        }
    }
    return tCalls;
}
/**
 * Find all t function calls in a class component.
 * @param path node path to the class component.
 */
function findTFunctionCallsInClassComponent(path) {
    const result = Array();
    const thisVisitor = {
        ThisExpression(path) {
            if (!path.parentPath.isMemberExpression())
                return;
            const propProperty = path.parentPath.get('property');
            if (Array.isArray(propProperty) || !propProperty.isIdentifier())
                return;
            if (propProperty.node.name !== 'props')
                return;
            // Ok, this is interesting, we have something with "this.props"
            if (path.parentPath.parentPath.isMemberExpression()) {
                // We have something in the form "this.props.xxxx".
                const tIdentifier = path.parentPath.parentPath.get('property');
                if (Array.isArray(tIdentifier) || !tIdentifier.isIdentifier())
                    return;
                if (tIdentifier.node.name !== 't')
                    return;
                // We have something in the form "this.props.t". Let's see if it's an
                // actual function call or an assignment.
                const tExpression = path.parentPath.parentPath.parentPath;
                if (tExpression.isCallExpression()) {
                    // Simple case. Direct call to "this.props.t()"
                    result.push(tExpression);
                }
                else if (tExpression.isVariableDeclarator()) {
                    // Hard case. const t = this.props.t;
                    // Let's loop through all references to t.
                    const id = tExpression.get('id');
                    if (!id.isIdentifier())
                        return;
                    for (const reference of id.scope.bindings[id.node.name]
                        .referencePaths) {
                        if (reference.parentPath.isCallExpression()) {
                            result.push(reference.parentPath);
                        }
                    }
                }
            }
            else if (path.parentPath.parentPath.isVariableDeclarator()) {
                // We have something in the form "const props = this.props"
                // Or "const {t} = this.props"
                const id = path.parentPath.parentPath.get('id');
                result.push(...findTFunctionCallsFromPropsAssignment(id));
            }
        },
    };
    path.traverse(thisVisitor);
    return result;
}
/**
 * Find t function calls in a function component.
 * @param path node path to the function component.
 */
function findTFunctionCallsInFunctionComponent(path) {
    const propsParam = path.get('params')[0];
    if (propsParam === undefined)
        return [];
    return findTFunctionCallsFromPropsAssignment(propsParam);
}
/**
 * Parse function or class declaration (likely components) to find whether
 * they are wrapped with "withTranslation()" HOC, and if so, extract all the
 * translations that come from the "t" function injected in the component
 * properties.
 *
 * @param path node path to the component
 * @param config plugin configuration
 * @param commentHints parsed comment hints
 */
function extractWithTranslationHOC(path, config, commentHints = []) {
    // Detect if this component is wrapped with withTranslation() somewhere
    const withTranslationCallExpression = findWithTranslationHOCCallExpression(path);
    if (withTranslationCallExpression === null)
        return [];
    let tCalls;
    if (path.isClassDeclaration()) {
        tCalls = findTFunctionCallsInClassComponent(path);
    }
    else {
        tCalls = findTFunctionCallsInFunctionComponent(path);
    }
    // Extract namespace
    let ns;
    const nsCommentHint = getCommentHintForPath(withTranslationCallExpression, 'NAMESPACE', commentHints);
    if (nsCommentHint) {
        // We got a comment hint, take its value as namespace.
        ns = nsCommentHint.value;
    }
    else {
        // Otherwise, try to get namespace from arguments.
        const namespaceArgument = withTranslationCallExpression.get('arguments')[0];
        ns = getFirstOrNull(evaluateIfConfident(namespaceArgument));
    }
    let keys = Array();
    for (const tCall of tCalls) {
        keys = [
            ...keys,
            ...extractTFunction(tCall, config, commentHints, true).map(k => ({
                // Add namespace if it was not explicitely set in t() call.
                ...k,
                parsedOptions: {
                    ...k.parsedOptions,
                    ns: k.parsedOptions.ns || ns,
                },
            })),
        ];
    }
    return keys;
}

/**
 * Handle the extraction.
 *
 * In case of ExtractionError occurring in the callback, a useful error
 * message will display and extraction will continue.
 *
 * @param path Current node path.
 * @param state Current visitor state.
 * @param callback Function to call that may throw ExtractionError.
 */
function handleExtraction(path, state, callback) {
    const filename = (state.file && state.file.opts.filename) || '???';
    const lineNumber = (path.node.loc && path.node.loc.start.line) || '???';
    const extractState = state.I18NextExtract;
    const collect = (keys) => {
        for (const key of keys) {
            if (extractState.extractedNodes.has(key.nodePath.node)) {
                // The node was already extracted. Skip it.
                continue;
            }
            extractState.extractedNodes.add(key.nodePath.node);
            state.I18NextExtract.extractedKeys.push(key);
        }
    };
    try {
        return callback(collect);
    }
    catch (err) {
        if (!(err instanceof ExtractionError)) {
            throw err;
        }
        // eslint-disable-next-line no-console
        console.warn(`${PLUGIN_NAME}: Extraction error in ${filename} at line ` +
            `${lineNumber}. ${err.message}`);
    }
}
const Visitor = {
    CallExpression(path, state) {
        const extractState = this.I18NextExtract;
        handleExtraction(path, state, collect => {
            collect(extractUseTranslationHook(path, extractState.config, extractState.commentHints));
            collect(extractTFunction(path, extractState.config, extractState.commentHints));
        });
    },
    JSXElement(path, state) {
        const extractState = this.I18NextExtract;
        handleExtraction(path, state, collect => {
            collect(extractTranslationRenderProp(path, extractState.config, extractState.commentHints));
            collect(extractTransComponent(path, extractState.config, extractState.commentHints));
        });
    },
    ClassDeclaration(path, state) {
        const extractState = this.I18NextExtract;
        handleExtraction(path, state, collect => {
            collect(extractWithTranslationHOC(path, extractState.config, extractState.commentHints));
        });
    },
    Function(path, state) {
        const extractState = this.I18NextExtract;
        handleExtraction(path, state, collect => {
            collect(extractWithTranslationHOC(path, extractState.config, extractState.commentHints));
        });
    },
};
function plugin (api) {
    api.assertVersion(7);
    // We have to store which nodes were extracted because the visitor might be
    // called multiple times by Babel and the state would be lost across calls.
    const extractedNodes = new WeakSet();
    // This is a cache for the exporter to keep track of the translation files.
    // It must remain global and persist across transpiled files.
    const exporterCache = createExporterCache();
    return {
        pre() {
            this.I18NextExtract = {
                config: parseConfig(this.opts),
                extractedKeys: [],
                commentHints: [],
                extractedNodes,
                exporterCache,
            };
        },
        post() {
            const extractState = this.I18NextExtract;
            if (extractState.extractedKeys.length === 0)
                return;
            for (const locale of extractState.config.locales) {
                const derivedKeys = extractState.extractedKeys.reduce((accumulator, k) => [
                    ...accumulator,
                    ...computeDerivedKeys(k, locale, extractState.config),
                ], Array());
                exportTranslationKeys(derivedKeys, locale, extractState.config, extractState.exporterCache);
            }
        },
        visitor: {
            Program(path, state) {
                // FIXME can't put this in Visitor because `path.traverse()` on a
                // Program node doesn't call the visitor for Program node.
                if (BabelTypes.isFile(path.container)) {
                    this.I18NextExtract.commentHints = parseCommentHints(path.container.comments);
                }
                path.traverse(Visitor, state);
            },
        },
    };
}

i18next.init();

module.exports = plugin;
